#==============================================================================================================================
# Title: Development of a machine learning model for prediction of GPCR-ligand interactions.
# Author: Mike Killoran, mike.killoran@promega.com
# Copyright 2021 Promega Corporation.
#==============================================================================================================================


#==============================================================================================================================
# Load Packages
#==============================================================================================================================


# allocate memory for large ML training function
options(java.parameters = "-Xmx8000m")

library("Rcpi")       # version 1.20.1
library("caret")      # version 6.0-86
library("mlr")        # version 2.17.1
library("doSNOW")     # version 1.0.18
library("tidyverse")  # version 1.3.0


#==============================================================================================================================
# Import data 
#==============================================================================================================================


setwd(">>PATH TO DIRECTORY<<")

# import data as CSV
glass.data <- read.csv(">>PATH TO FILE<</glass_dataset.csv", header = TRUE, as.is = TRUE)


#==============================================================================================================================
# Pre-process data 
#==============================================================================================================================

# filter for unique SMILES-family combinations
filtered.data <- glass.data %>% group_by(smiles, family) %>% filter(row_number() == 1)

# determine families that have less than 700 molecules
fams.low <- as.data.frame(table(glass.data$family)) %>% arrange(Freq) %>% filter(Freq < 700) %>% pull(Var1) %>% as.vector

# select molecules from filtered dataset that in families above threshold
filtered.data.high <- filtered.data %>% filter(!family %in% fams.low)

# random subset 1
set.seed(54321)
selected.data <- filtered.data.high %>% group_by(family) %>% sample_n(530, replace = FALSE)


#==============================================================================================================================
# Calculate molecular fingerprints
#==============================================================================================================================

# go through script one subset at a time
selected.smiles <- selected.data$smiles

# set loop parameters based on number of elements in drugseq
a <- floor((length(selected.smiles)/10))
b <- (length(selected.smiles) %% 10)  

# preallocate container vectors of the appropriate size to fill with lists
fp.list.1 <- vector('list',length(selected.smiles))
fp.list.2 <- vector('list',length(selected.smiles))
fp.list.3 <- vector('list',length(selected.smiles))


# calculate molecular fingerprints and add to list
for (i in 1:a) {
  fp.list.1[[i]] <- data.frame(extractDrugEstateComplete(readMolFromSmi(textConnection(selected.smiles[((i*10)-9):(i*10)]), type = 'mol'))) }


for (i in 1:a) {
  fp.list.2[[i]] <- data.frame(extractDrugMACCSComplete(readMolFromSmi(textConnection(selected.smiles[((i*10)-9):(i*10)]), type = 'mol'))) }


for (i in 1:a) {
  fp.list.3[[i]] <- data.frame(extractDrugOBFP4(selected.smiles[((i*10)-9):(i*10)], type = 'smile')) }


# gives error when a / 10 has no remainder and b = 0, can be ignored
for (i in (a * 10 + 1) : (a * 10 + 1) ) {
  fp.list.1[[i]] <- data.frame(extractDrugEstateComplete(readMolFromSmi(textConnection(selected.smiles[i : (a*10 + b)]), type = 'mol')))
  fp.list.2[[i]] <- data.frame(extractDrugMACCSComplete(readMolFromSmi(textConnection(selected.smiles[i : (a*10 + b)]), type = 'mol')))
  fp.list.3[[i]] <- data.frame(extractDrugOBFP4(selected.smiles[i : (a*10 + b)], type = 'smile'))
}

# collapse lists into single dataframe
fp.df.1 <- do.call('rbind', fp.list.1)
fp.df.2 <- do.call('rbind', fp.list.2)
fp.df.3 <- do.call('rbind', fp.list.3)

# rename columns to unique names before merge
fp.df.1 %>% data.table::setnames(old = fp.df.1 %>% names(), new = fp.df.1 %>% names() %>% paste0("_EC"))
fp.df.2 %>% data.table::setnames(old = fp.df.2 %>% names(), new = fp.df.2 %>% names() %>% paste0("_MACCS"))
fp.df.3 %>% data.table::setnames(old = fp.df.3 %>% names(), new = fp.df.3 %>% names() %>% paste0("_OBFP4"))


#==============================================================================================================================
# Combine into single dataframe and add family labels
#==============================================================================================================================

# combine all columns into single dataframe of fingerprints
fp.df <- as.matrix(cbind(selected.smiles, fp.df.1, fp.df.2, fp.df.3))
fp.df <- as.matrix(cbind(fp.df.1, fp.df.2, fp.df.3))

# remove zero variance descriptors
fp.df.nzv = as.data.frame(fp.df[, -nearZeroVar(fp.df)])

# combine family labels with fingerprints
fp.data <- cbind('family' = selected.data$family, fp.df.nzv)

# import a table to convert family names to letter IDs for brevity
fam_to_ID <- read.csv(">>PATH TO FILE<<", header = TRUE, as.is = TRUE)

# merge the datasets to connect family names with letter IDs
fp.data <- as.data.frame(merge(x = fp.data, y = fam_to_ID, by = "family", all.x = TRUE))

# remove unneeded variables
rm(fam_to_ID, filtered.data, filtered.data.high, fp.df, fp.df.1, fp.df.2, fp.df.3, fp.df.nzv, fp.list.1, fp.list.2, fp.list.3, glass.data, selected.data, a, b, fams.low, i, selected.smiles)

#==============================================================================================================================
# Split Data into training and test sets
#==============================================================================================================================

# create indexes for splitting data 70/30
set.seed(54321) 
indexes <- createDataPartition(fp.data$family,
                                 times = 1,
                                 p = 0.7,
                                 list = FALSE)

# use indexes to subset data into training and test sets
train.data <- fp.data[indexes,]
test.data <- fp.data[-indexes,]


#===============================================================================================================================
# Train the random forest classification model  
#===============================================================================================================================

# started here on 7/31/2020
#load("C:/Users/mkillora/Desktop/20190213_adrenergic_split_results.RData")


# cross-validation settings
ctrl = trainControl(method = "repeatedcv", number = 5, repeats = 10, classProbs = TRUE)

# start cluster for parallel training
cl <- makeCluster(5, type = "SOCK")
registerDoSNOW(cl)

# train a random forest classifier
set.seed(54321)
rf.fit = caret::train(family ~ .,
               data = train.data,
               method = "rf", 
               trControl = ctrl,
               metric = "Accuracy", 
               preProc = c("center", "scale"))

# end cluster
stopCluster(cl)

# print the cross-validation and model results
print(rf.fit)
print(rf.fit$finalModel)

# save data environment
save.image(file='FILE NAME.RData')

# save the model
setwd(">>PATH TO DIRECTORY<<")
saveRDS(rf.fit, "FILE NAME.rds")


#===============================================================================================================================
# Predict with test set    
#===============================================================================================================================

# Make predictions on the test set using a model trained on all rows of the training set using the found optimal hyperparameter values.
preds <- predict(rf.fit, test.data)
preds_prob <- predict(rf.fit, test.data, type = "prob")

# calculate confusion matrix for test set predictions
cm <- confusionMatrix(preds, test.data$family) # need to change test set outcome to a factor type

# output table and confusion matrix
cm.tab <- as.table(cm)
cm.mat <- as.matrix(cm, what = "classes")


#===============================================================================================================================
# Predict with validation set    
#===============================================================================================================================

# load the validation set of molecules as a list of SMILES in CSV format
val.set.0 = read.csv(">>PATH TO FILE<<", header = TRUE, as.is = TRUE)
val.set = read.csv(">>PATH TO FILE<<", header = TRUE, as.is = TRUE)[2]

outcomes <- fam_conv

# look at the data
head(val.set)

# convert to character vector
valseq = as.character(unlist(val.set))
valseq[1:5]

# Extract molecular descriptors for molecules in validation set
v.df.1 <- data.frame(extractDrugEstateComplete(readMolFromSmi(textConnection(valseq), type = 'mol')))
v.df.2 <- data.frame(extractDrugMACCSComplete(readMolFromSmi(textConnection(valseq), type = 'mol')))
v.df.3 <- data.frame(extractDrugOBFP4((valseq), type = 'smile'))

# rename columns to preserve suffix
v.df.1 %>% data.table::setnames(old = v.df.1 %>% names(), new = v.df.1 %>% names() %>% paste0("_EC"))
v.df.2 %>% data.table::setnames(old = v.df.2 %>% names(), new = v.df.2 %>% names() %>% paste0("_MACCS"))
v.df.3 %>% data.table::setnames(old = v.df.3 %>% names(), new = v.df.3 %>% names() %>% paste0("_OBFP4"))

# merge to single df
v.df <- cbind(v.df.1, v.df.2, v.df.3)

# make predictions, preds can be done with data lacking the response variable column
preds.val <- predict(rf.fit, v.df)
preds.val.prob <- predict(rf.fit, v.df, type = "prob")


a <- preds.val.prob
rownames(a) <- val.set.0$molecule
a <- as.data.frame(t(a))
a <- a * 100
a <- tibble::rownames_to_column(a, "outcome")
a <- merge(x = outcomes, y = a, by = "outcome")



